/*
 * set_empeg_max_fid.c by Mark Lord, March 2005.
 *
 * Permission hereby granted to use, modify, and/or redistribute this in full,
 * so long as full source code is included and this notice remains intact.
 *
 * Version 1: initial release.
 * Version 2: fixed year (above), and boundary on search.
 * Version 3: multiply sector count by 2.
 * Version 4: print version; subtract dynamic playlist overhead (0x1000) from total sectors.
 */

static const int my_version = 4;

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <string.h>
#include <errno.h>

#ifndef MAP_FAILED
#define MAP_FAILED ((void *)-1UL)
#endif

static void hexdump (const char *prefix, void *p, unsigned int offset, int len)
{
	printf("%s: %08x:", prefix, offset);
	while (len--) {
		printf(" %02x", *(unsigned char *)p);
		++p;
	}
	printf("\n");
}

static void err_exit (const char *s)
{
	int err = errno;
	perror(s);
	exit(err);
}

static unsigned int get_hda3_sectors (void)
{
	int fd;
	unsigned int len, dont_care;
	char *pathname = "/proc/partitions", buf[1024], *p;

	fd = open(pathname, O_RDONLY);
	if (fd == -1)
		err_exit(pathname);
	len = read(fd, buf, sizeof(buf));
	if (len < 1)
		err_exit(pathname);
	close(fd);
	buf[len] = '\0';
	len = -1;
	p = strstr(buf, "hda2");
	if (!p || 2 != sscanf(p, "hda2 3 3 %u hda3 %u", &len, &dont_care)) {
		fprintf(stderr, "failed to parse %s\n", pathname);
		exit(1);
	}
	printf("/dev/hda3: %d sectors\n", len);
	return len;
}

static void store_le32 (unsigned char *p, unsigned int val)
{
	p[0] =  val        & 0xff;
	p[1] = (val >>  8) & 0xff;
	p[2] = (val >> 16) & 0xff;
	p[3] = (val >> 24) & 0xff;
}

static int compare_le32 (unsigned char *p, unsigned int val)
{
	return
	   (p[0] ==  (val        & 0xff)
	 && p[1] == ((val >>  8) & 0xff)
	 && p[2] == ((val >> 16) & 0xff)
	 && p[3] == ((val >> 24) & 0xff));
}

static void usage (const char *myname)
{
	fprintf(stderr, "Usage: %s  /empeg/bin/player  [max_fid]\n", myname);
	exit(1);
}

int main (int argc, char *argv[])
{
	int fd;
	struct stat st;
	char *pathname = "/empeg/bin/player";
	void *start, *end, *p, *found = NULL;
	int max_fid = -1;

	printf("%s: version %u by Mark Lord (March 2005)\n", argv[0], my_version);

	if (argc > 1) { 
		pathname = argv[1];
		if (argc > 3 || pathname[0] != '/' || !pathname[1])
			usage(argv[0]);
		if (argc == 3 && 1 != sscanf(argv[2], "%d", &max_fid))
			usage(argv[0]);
	}
	fd = open(pathname, O_RDWR);
	if (fd == -1)
		err_exit(pathname);
	if (-1 == fstat(fd, &st))
		err_exit(pathname);
	start = mmap(NULL, st.st_size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	if (start == MAP_FAILED)
		err_exit(pathname);
	end = start + (st.st_size & ~3) - 4;

	for (p = start; p != end; p += 4) {
		if (compare_le32(p, 0x00001000) && compare_le32(p+4, 0x00007000)) {
			if (found) {
				fprintf(stderr, "Huh? found multiple patch points (%08x, %08x)\n",
								p - start, found - start);
				exit(1);
			}
			found = p;
		}
	}
	if (!found) {
		fprintf(stderr, "Huh? patch point not found (already patched?)\n");
		exit(1);
	}

	if (argc != 3)
		max_fid = (2 * get_hda3_sectors()) - 0x1000;
	printf("New max_fid value will be %d (%04x %04x)\n", max_fid, max_fid & 0xffff, max_fid >> 16);
	if (max_fid < 0) {
		fprintf(stderr, "huh? cannot use negative max_fid value\n");
		exit(1);
	}

	hexdump("before", found, found - start, 16);
	store_le32(found+4, max_fid);	/* patch it */
	hexdump(" after", found, found - start, 16);

	munmap(start, st.st_size);
	close(fd);
	exit(0);
}
